class CommandWrappers(object):
    def version(self):
        """Retrieve daemon and system specific version information.

        :return: daemon and system specific version information
        :rtype: dict
        """
        return self.request("version")

    def stats(self):
        """Retrieve IKE daemon statistics and load information.

        :return: IKE daemon statistics and load information
        :rtype: dict
        """
        return self.request("stats")

    def reload_settings(self):
        """Reload strongswan.conf settings and any plugins supporting reload.
        """
        self.request("reload-settings")

    def initiate(self, sa):
        """Initiate an SA.

        :param sa: the SA to initiate
        :type sa: dict
        :return: generator for logs emitted as dict
        :rtype: generator
        """
        return self.streamed_request("initiate", "control-log", sa)

    def terminate(self, sa):
        """Terminate an SA.

        :param sa: the SA to terminate
        :type sa: dict
        :return: generator for logs emitted as dict
        :rtype: generator
        """
        return self.streamed_request("terminate", "control-log", sa)

    def rekey(self, sa):
        """Initiate the rekeying of an SA.

        .. versionadded:: 5.5.2

        :param sa: the SA to rekey
        :type sa: dict
        :return: number of matched SAs
        :rtype: dict
        """
        return self.request("rekey", sa)

    def redirect(self, sa):
        """Redirect an IKE_SA.

        .. versionchanged:: 5.5.2
           The number of matched SAs is returned.

        :param sa: the SA to redirect
        :type sa: dict
        :return: number of matched SAs
        :rtype: dict
        """
        return self.request("redirect", sa)

    def install(self, policy):
        """Install a trap, drop or bypass policy defined by a CHILD_SA config.

        :param policy: policy to install
        :type policy: dict
        """
        self.request("install", policy)

    def uninstall(self, policy):
        """Uninstall a trap, drop or bypass policy defined by a CHILD_SA config.

        :param policy: policy to uninstall
        :type policy: dict
        """
        self.request("uninstall", policy)

    def list_sas(self, filters=None):
        """Retrieve active IKE_SAs and associated CHILD_SAs.

        :param filters: retrieve only matching IKE_SAs (optional)
        :type filters: dict
        :return: generator for active IKE_SAs and associated CHILD_SAs as dict
        :rtype: generator
        """
        return self.streamed_request("list-sas", "list-sa", filters)

    def list_policies(self, filters=None):
        """Retrieve installed trap, drop and bypass policies.

        :param filters: retrieve only matching policies (optional)
        :type filters: dict
        :return: generator for installed trap, drop and bypass policies as dict
        :rtype: generator
        """
        return self.streamed_request("list-policies", "list-policy",
                                     filters)

    def list_conns(self, filters=None):
        """Retrieve loaded connections.

        :param filters: retrieve only matching configuration names (optional)
        :type filters: dict
        :return: generator for loaded connections as dict
        :rtype: generator
        """
        return self.streamed_request("list-conns", "list-conn",
                                     filters)

    def get_conns(self):
        """Retrieve connection names loaded exclusively over vici.

        :return: connection names
        :rtype: dict
        """
        return self.request("get-conns")

    def list_certs(self, filters=None):
        """Retrieve loaded certificates.

        :param filters: retrieve only matching certificates (optional)
        :type filters: dict
        :return: generator for loaded certificates as dict
        :rtype: generator
        """
        return self.streamed_request("list-certs", "list-cert", filters)

    def list_authorities(self, filters=None):
        """Retrieve loaded certification authority information.

        .. versionadded:: 5.3.3

        :param filters: retrieve only matching CAs (optional)
        :type filters: dict
        :return: generator for loaded CAs as dict
        :rtype: generator
        """
        return self.streamed_request("list-authorities", "list-authority",
                                     filters)

    def get_authorities(self):
        """Retrieve certification authority names loaded exclusively over vici.

        :return: CA names
        :rtype: dict
        """
        return self.request("get-authorities")

    def load_conn(self, connection):
        """Load a connection definition into the daemon.

        :param connection: connection definition
        :type connection: dict
        """
        self.request("load-conn", connection)

    def unload_conn(self, name):
        """Unload a connection definition.

        :param name: connection definition name
        :type name: dict
        """
        self.request("unload-conn", name)

    def load_cert(self, certificate):
        """Load a certificate into the daemon.

        :param certificate: PEM or DER encoded certificate
        :type certificate: dict
        """
        self.request("load-cert", certificate)

    def load_key(self, private_key):
        """Load a private key into the daemon.

        .. versionchanged:: 5.5.3
           The key identifier of the loaded key is returned.

        :param private_key: PEM or DER encoded key
        :type private_key: dict
        :return: key identifier
        :rtype: dict
        """
        return self.request("load-key", private_key)

    def unload_key(self, key_id):
        """Unload the private key with the given key identifier.

        .. versionadded:: 5.5.2

        :param key_id: key identifier
        :type key_id: dict
        """
        self.request("unload-key", key_id)

    def get_keys(self):
        """Retrieve identifiers of private keys loaded exclusively over vici.

        .. versionadded:: 5.5.2

        :return: key identifiers
        :rtype: dict
        """
        return self.request("get-keys")

    def load_token(self, token):
        """Load a private key located on a token into the daemon.

        .. versionadded:: 5.5.2

        :param token: token details
        :type token: dict
        :return: key identifier
        :rtype: dict
        """
        return self.request("load-token", token)

    def load_shared(self, secret):
        """Load a shared IKE PSK, EAP or XAuth secret into the daemon.

        .. versionchanged:: 5.5.2
           A unique identifier may be associated with the secret.

        :param secret: shared IKE PSK, EAP or XAuth secret
        :type secret: dict
        """
        self.request("load-shared", secret)

    def unload_shared(self, identifier):
        """Unload a previously loaded shared secret by its unique identifier.

        .. versionadded:: 5.5.2

        :param identifier: unique identifier
        :type secret: dict
        """
        self.request("unload-shared", identifier)

    def get_shared(self):
        """Retrieve identifiers of shared keys loaded exclusively over vici.

        .. versionadded:: 5.5.2

        :return: identifiers
        :rtype: dict
        """
        return self.request("get-shared")

    def flush_certs(self, filter=None):
        """Flush the volatile certificate cache.

        Flush the certificate stored temporarily in the cache. The filter
        allows to flush only a certain type of certificates, e.g. CRLs.

        :param filter: flush only certificates of a given type (optional)
        :type filter: dict
        """
        self.request("flush-certs", filter)

    def clear_creds(self):
        """Clear credentials loaded over vici.

        Clear all loaded certificate, private key and shared key credentials.
        This affects only credentials loaded over vici, but additionally
        flushes the credential cache.
        """
        self.request("clear-creds")

    def load_authority(self, ca):
        """Load a certification authority definition into the daemon.

        :param ca: certification authority definition
        :type ca: dict
        """
        self.request("load-authority", ca)

    def unload_authority(self, ca):
        """Unload a previously loaded certification authority by name.

        :param ca: certification authority name
        :type ca: dict
        """
        self.request("unload-authority", ca)

    def load_pool(self, pool):
        """Load a virtual IP pool.

        Load an in-memory virtual IP and configuration attribute pool.
        Existing pools with the same name get updated, if possible.

        :param pool: virtual IP and configuration attribute pool
        :type pool: dict
        """
        return self.request("load-pool", pool)

    def unload_pool(self, pool_name):
        """Unload a virtual IP pool.

        Unload a previously loaded virtual IP and configuration attribute pool.
        Unloading fails for pools with leases currently online.

        :param pool_name: pool by name
        :type pool_name: dict
        """
        self.request("unload-pool", pool_name)

    def get_pools(self, options=None):
        """Retrieve loaded pools.

        :param options: filter by name and/or retrieve leases (optional)
        :type options: dict
        :return: loaded pools
        :rtype: dict
        """
        return self.request("get-pools", options)

    def get_algorithms(self):
        """List of currently loaded algorithms and their implementation.

        .. versionadded:: 5.4.0

        :return: algorithms
        :rtype: dict
        """
        return self.request("get-algorithms")

    def get_counters(self, options=None):
        """List global or connection-specific counters for several IKE events.

        .. versionadded:: 5.6.1

        :param options: get global counters or those of all or one connection
        :type options: dict
        :return: counters
        :rtype: dict
        """
        return self.request("get-counters", options)

    def reset_counters(self, options=None):
        """Reset global or connection-specific IKE event counters.

        .. versionadded:: 5.6.1

        :param options: reset global counters or those of all or one connection
        :type options: dict
        """
        self.request("reset-counters", options)



from functools import wraps
import inspect

from .protocol import RECV_TIMEOUT_DEFAULT


class StopListening(Exception):
    """Exception that may be raised to stop listening for events."""


class EventListener(object):
    def __init__(self, session=None):
        """Create an event listener instance, which provides decorator methods
        to make listening for events and the disconnection of the vici session
        more convenient.

        The session is optional here, but one must be set via
        :func:`~set_session()` before calling :func:`~listen()`.

        :param session: optional vici session to use
        :type session: :class:`~vici.session.Session` or None
        """
        self.event_map = {}
        self.disconnect_list = []
        self.timeout_list = []
        self.session = session

    def set_session(self, session):
        """Set the session that's used to listen for events. Only has an effect
        when set before calling :func:`~listen()`.

        :param session: vici session to use
        :type session: :class:`~vici.session.Session`
        """
        self.session = session

    def on_events(self, events):
        """Decorator to mark a function as a listener for specific events.

        The decorated function is expected to receive the name of the event and
        the data as arguments. It may raise :class:`~StopListening` to stop
        listening and let :func:`~listen()` return.

        :param events: events to register and call decorated function for
        :type events: list
        :return: decorator function
        :rtype: any
        """
        def decorator(func):
            self.event_map.update({event: func for event in events})

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def on_disconnected(self):
        """Decorator to mark a function as a listener for when the daemon
        disconnects the vici session.

        This listener instance is passed to the decorated function, which may
        be used to set a new session and continue listening. If no session is
        set, :func:`~listen()` will return after the decorated function has
        been called.

        :return: decorator function
        :rtype: any
        """
        def decorator(func):
            self.disconnect_list.append(func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def on_timeout(self):
        """Decorator to mark a function as a listener for when a timeout occurs
        while waiting for events. Only has an effect if :func:`~listen()` is
        called with a timeout.

        The decorated function may either take no or two arguments (both will be
        set to `None`). So this may be applied to a function that's also
        decorated with :func:`~on_events()`. It may raise
        :class:`~StopListening` to stop listening and let :func:`~listen()`
        return.

        :return: decorator function
        :rtype: any
        """
        def decorator(func):
            self.timeout_list.append(func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def listen(self, timeout=RECV_TIMEOUT_DEFAULT):
        """Dispatch events registered via decorators of this instance.

        An active session has to be set before calling this. After getting
        disconnected, a new session may be set via :func:`~set_session()` in
        a function decorated with :func:`~on_disconnected()` to resume
        listening for events.

        This method does not return unless :class:`~StopListening` or an
        unexpected exception is raised or if the current session is disconnected
        and no new session is set in a listener.

        The optional timeout allows calling functions decorated with
        :func:`~on_timeout()` if no event has been received for that time. Which
        may be used to abort listening or perform periodic tasks while
        continuing to listen for events.

        :param timeout: timeout to wait for events, in fractions of a second
        :type timeout: float
        :return: True if StopListening was raised, False if no session available
        :rtype: bool
        """
        while True:
            try:
                if self.session is None:
                    return False
                for label, event in self.session.listen(self.event_map.keys(),
                                                        timeout):
                    if label is None and event is None:
                        for func in self.timeout_list:
                            if len(inspect.signature(func).parameters) > 0:
                                func(label, event)
                            else:
                                func()
                        continue
                    name = label.decode()
                    if name in self.event_map:
                        self.event_map[name](name, event)
            except IOError:
                self.session = None
                for func in self.disconnect_list:
                    func(self)
                continue
            except StopListening:
                return True



import socket
import platform

from .exception import SessionException, CommandException, EventUnknownException
from .protocol import Transport, Packet, Message, RECV_TIMEOUT_DEFAULT
from .command_wrappers import CommandWrappers


class Session(CommandWrappers, object):
    def __init__(self, sock=None):
        """Establish a session with an IKE daemon.

        By default, the session will connect to the `/var/run/charon.vici` Unix
        domain socket.

        If there is a need to connect a socket in another location or set
        specific settings on the socket (like a timeout), create and connect
        a socket and pass it to the `sock` parameter.

        .. note::

            In case a timeout is set on the socket, the internal read code
            will temporarily disable it after receiving the first byte to avoid
            partial read corruptions.

        :param sock: socket connected to the IKE daemon (optional)
        :type sock: socket.socket
        """
        if sock is None:
            if platform.system() == "Windows":
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect(('127.0.0.1', 4502))
            else:
                sock = socket.socket(socket.AF_UNIX)
                sock.connect("/var/run/charon.vici")
        self.transport = Transport(sock)

    def _communicate(self, packet):
        """Send packet over transport and parse response.

        :param packet: packet to send
        :type packet: :py:class:`vici.protocol.Packet`
        :return: parsed packet in a tuple with message type and payload
        :rtype: :py:class:`collections.namedtuple`
        """
        self.transport.send(packet)
        return Packet.parse(self.transport.receive())

    def _register_unregister(self, event_type, register):
        """Register or unregister for the given event.

        :param event_type: event to register
        :type event_type: str
        :param register: whether to register or unregister
        :type register: bool
        """
        if register:
            packet = Packet.register_event(event_type)
        else:
            packet = Packet.unregister_event(event_type)
        response = self._communicate(packet)
        if response.response_type == Packet.EVENT_UNKNOWN:
            raise EventUnknownException(
                "Unknown event type '{event}'".format(event=event_type)
            )
        while response.response_type == Packet.EVENT:
            response = Packet.parse(self.transport.receive())
        if response.response_type != Packet.EVENT_CONFIRM:
            raise SessionException(
                "Unexpected response type {type}, "
                "expected '{confirm}' (EVENT_CONFIRM)".format(
                    type=response.response_type,
                    confirm=Packet.EVENT_CONFIRM,
                )
            )

    def request(self, command, message=None):
        """Send request with an optional message.

        :param command: command to send
        :type command: str
        :param message: message (optional)
        :type message: str
        :return: command result
        :rtype: dict
        """
        if message is not None:
            message = Message.serialize(message)
        packet = Packet.request(command, message)
        response = self._communicate(packet)

        if response.response_type != Packet.CMD_RESPONSE:
            raise SessionException(
                "Unexpected response type {type}, "
                "expected '{response}' (CMD_RESPONSE)".format(
                    type=response.response_type,
                    response=Packet.CMD_RESPONSE
                )
            )

        command_response = Message.deserialize(response.payload)
        if "success" in command_response:
            if command_response["success"] != b"yes":
                raise CommandException(
                    "Command failed: {errmsg}".format(
                        errmsg=command_response["errmsg"].decode("UTF-8")
                    )
                )

        return command_response

    def streamed_request(self, command, event_stream_type, message=None):
        """Send command request and collect and return all emitted events.

        :param command: command to send
        :type command: str
        :param event_stream_type: event type emitted on command execution
        :type event_stream_type: str
        :param message: message (optional)
        :type message: str
        :return: generator for streamed event responses as dict
        :rtype: generator
        """
        if message is not None:
            message = Message.serialize(message)

        self._register_unregister(event_stream_type, True)

        try:
            packet = Packet.request(command, message)
            self.transport.send(packet)
            exited = False
            while True:
                response = Packet.parse(self.transport.receive())
                if response.response_type == Packet.EVENT:
                    if not exited:
                        try:
                            yield Message.deserialize(response.payload)
                        except GeneratorExit:
                            exited = True
                else:
                    break

            if response.response_type == Packet.CMD_RESPONSE:
                command_response = Message.deserialize(response.payload)
            else:
                raise SessionException(
                    "Unexpected response type {type}, "
                    "expected '{response}' (CMD_RESPONSE)".format(
                        type=response.response_type,
                        response=Packet.CMD_RESPONSE
                    )
                )

        finally:
            self._register_unregister(event_stream_type, False)

        # evaluate command result, if any
        if "success" in command_response:
            if command_response["success"] != b"yes":
                raise CommandException(
                    "Command failed: {errmsg}".format(
                        errmsg=command_response["errmsg"].decode("UTF-8")
                    )
                )

    def listen(self, event_types, timeout=RECV_TIMEOUT_DEFAULT):
        """Register and listen for the given events.

        If a timeout is given, the generator produces a (None, None) tuple
        if no event has been received for that time. This allows the caller
        to either abort by breaking from the generator, or perform periodic
        tasks while staying registered within listen(), and then continue
        waiting for more events.

        :param event_types: event types to register
        :type event_types: list
        :param timeout: timeout to wait for events, in fractions of a second
        :type timeout: float
        :return: generator for streamed event responses as (event_type, dict)
        :rtype: generator
        """
        for event_type in event_types:
            self._register_unregister(event_type, True)

        try:
            while True:
                try:
                    response = Packet.parse(self.transport.receive(timeout))
                except socket.timeout:
                    yield None, None
                    continue
                if response.response_type == Packet.EVENT:
                    try:
                        msg = Message.deserialize(response.payload)
                        yield response.event_type, msg
                    except GeneratorExit:
                        break

        finally:
            for event_type in event_types:
                self._register_unregister(event_type, False)



"""Exception types that may be thrown by this library."""


class DeserializationException(Exception):
    """Encountered an unexpected byte sequence or missing element type."""


class SessionException(Exception):
    """Session request exception."""


class CommandException(Exception):
    """Command result exception."""


class EventUnknownException(Exception):
    """Event unknown exception."""



import io
import socket
import struct

from collections import namedtuple
from collections import OrderedDict

from .exception import DeserializationException


RECV_TIMEOUT_DEFAULT = object()


class Transport(object):
    HEADER_LENGTH = 4
    MAX_SEGMENT = 512 * 1024

    def __init__(self, sock):
        self.socket = sock

    def send(self, packet):
        self.socket.sendall(struct.pack("!I", len(packet)) + packet)

    def receive(self, timeout=RECV_TIMEOUT_DEFAULT):
        raw_length = self._recvall(self.HEADER_LENGTH, timeout)
        length, = struct.unpack("!I", raw_length)
        payload = self._recvall(length)
        return payload

    def close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def _recvall(self, count, timeout=RECV_TIMEOUT_DEFAULT):
        """Ensure to read count bytes from the socket"""
        data = b""
        old_timeout = self.socket.gettimeout()
        if timeout is not RECV_TIMEOUT_DEFAULT:
            self.socket.settimeout(timeout)
        try:
            while len(data) < count:
                buf = self.socket.recv(count - len(data))
                self.socket.settimeout(None)
                if not buf:
                    raise socket.error('Connection closed')
                data += buf
        finally:
            self.socket.settimeout(old_timeout)
        return data


class Packet(object):
    CMD_REQUEST = 0         # Named request message
    CMD_RESPONSE = 1        # Unnamed response message for a request
    CMD_UNKNOWN = 2         # Unnamed response if requested command is unknown
    EVENT_REGISTER = 3      # Named event registration request
    EVENT_UNREGISTER = 4    # Named event de-registration request
    EVENT_CONFIRM = 5       # Unnamed confirmation for event (de-)registration
    EVENT_UNKNOWN = 6       # Unnamed response if event (de-)registration failed
    EVENT = 7               # Named event message

    ParsedPacket = namedtuple(
        "ParsedPacket",
        ["response_type", "payload"]
    )

    ParsedEventPacket = namedtuple(
        "ParsedEventPacket",
        ["response_type", "event_type", "payload"]
    )

    @classmethod
    def _named_request(cls, request_type, request, message=None):
        request = request.encode("UTF-8")
        payload = struct.pack("!BB", request_type, len(request)) + request
        if message is not None:
            return payload + message
        else:
            return payload

    @classmethod
    def request(cls, command, message=None):
        return cls._named_request(cls.CMD_REQUEST, command, message)

    @classmethod
    def register_event(cls, event_type):
        return cls._named_request(cls.EVENT_REGISTER, event_type)

    @classmethod
    def unregister_event(cls, event_type):
        return cls._named_request(cls.EVENT_UNREGISTER, event_type)

    @classmethod
    def parse(cls, packet):
        stream = FiniteStream(packet)
        response_type, = struct.unpack("!B", stream.read(1))

        if response_type == cls.EVENT:
            length, = struct.unpack("!B", stream.read(1))
            event_type = stream.read(length)
            return cls.ParsedEventPacket(response_type, event_type, stream)
        else:
            return cls.ParsedPacket(response_type, stream)


class Message(object):
    SECTION_START = 1       # Begin a new section having a name
    SECTION_END = 2         # End a previously started section
    KEY_VALUE = 3           # Define a value for a named key in the section
    LIST_START = 4          # Begin a named list for list items
    LIST_ITEM = 5           # Define an unnamed item value in the current list
    LIST_END = 6            # End a previously started list

    @classmethod
    def serialize(cls, message):
        def encode_named_type(marker, name):
            name = name.encode("UTF-8")
            return struct.pack("!BB", marker, len(name)) + name

        def encode_blob(value):
            if not isinstance(value, bytes):
                value = str(value).encode("UTF-8")
            return struct.pack("!H", len(value)) + value

        def serialize_list(lst):
            segment = bytes()
            for item in lst:
                segment += struct.pack("!B", cls.LIST_ITEM) + encode_blob(item)
            return segment

        def serialize_dict(d):
            segment = bytes()
            for key, value in d.items():
                if isinstance(value, dict):
                    segment += (
                        encode_named_type(cls.SECTION_START, key)
                        + serialize_dict(value)
                        + struct.pack("!B", cls.SECTION_END)
                    )
                elif isinstance(value, list):
                    segment += (
                        encode_named_type(cls.LIST_START, key)
                        + serialize_list(value)
                        + struct.pack("!B", cls.LIST_END)
                    )
                else:
                    segment += (
                        encode_named_type(cls.KEY_VALUE, key)
                        + encode_blob(value)
                    )
            return segment

        return serialize_dict(message)

    @classmethod
    def deserialize(cls, stream):
        def decode_named_type(stream):
            length, = struct.unpack("!B", stream.read(1))
            return stream.read(length).decode("UTF-8")

        def decode_blob(stream):
            length, = struct.unpack("!H", stream.read(2))
            return stream.read(length)

        def decode_list_item(stream):
            marker, = struct.unpack("!B", stream.read(1))
            while marker == cls.LIST_ITEM:
                yield decode_blob(stream)
                marker, = struct.unpack("!B", stream.read(1))

            if marker != cls.LIST_END:
                raise DeserializationException(
                    "Expected end of list at {pos}".format(pos=stream.tell())
                )

        section = OrderedDict()
        section_stack = []
        while stream.has_more():
            element_type, = struct.unpack("!B", stream.read(1))
            if element_type == cls.SECTION_START:
                section_name = decode_named_type(stream)
                new_section = OrderedDict()
                section[section_name] = new_section
                section_stack.append(section)
                section = new_section

            elif element_type == cls.LIST_START:
                list_name = decode_named_type(stream)
                section[list_name] = [item for item in decode_list_item(stream)]

            elif element_type == cls.KEY_VALUE:
                key = decode_named_type(stream)
                section[key] = decode_blob(stream)

            elif element_type == cls.SECTION_END:
                if len(section_stack):
                    section = section_stack.pop()
                else:
                    raise DeserializationException(
                        "Unexpected end of section at {pos}".format(
                            pos=stream.tell()
                        )
                    )

        if len(section_stack):
            raise DeserializationException("Expected end of section")
        return section


class FiniteStream(io.BytesIO):
    def __len__(self):
        return len(self.getvalue())

    def has_more(self):
        return self.tell() < len(self)
